/*
 * Copyright 2022-2023 SPACEMIT. All rights reserved.
 * Use of this source code is governed by a BSD-style license
 * that can be found in the LICENSE file.
 *
 * @Author: David(qiang.fu@spacemit.com)
 * @Date: 2023-08-30 15:31:42
 * @LastEditTime: 2023-11-07 10:51:38
 * @Description:
 */

#include <sys/mman.h>

#include "libv4l-stmpp.h"

int stmpp_to_v4l2_buffer(struct stmpp_context *ctx,
                         struct stmpp_buffer *stmpp_buffer,
                         struct v4l2_buffer *buffer) {
  unsigned int i;

  ENTER();

  if (buffer->length != stmpp_buffer->length) {
    LOGE("wrong buffer planes: %d(expected: %d)\n", buffer->length,
         stmpp_buffer->length);
    return -1;
  }

  // copy data of each plane
  for (i = 0; i < buffer->length; i++) {
    buffer->m.planes[i].length = stmpp_buffer->planes[i].length;
    buffer->m.planes[i].data_offset = stmpp_buffer->planes[i].data_offset;
    buffer->m.planes[i].bytesused = 0;

    if (buffer->memory == V4L2_MEMORY_MMAP)
      /* Only support mem_offset for plane 0 */
      buffer->m.planes[i].m.mem_offset =
          i ? 0 : STMPP_MEM_OFFSET(buffer->type, buffer->index);
    else if (buffer->memory == V4L2_MEMORY_USERPTR)
      buffer->m.planes[i].m.userptr = stmpp_buffer->planes[i].userptr;
    else if (buffer->memory == V4L2_MEMORY_DMABUF)
      buffer->m.planes[i].m.fd = stmpp_buffer->planes[i].fd;
  }

  if (stmpp_buffer_available(stmpp_buffer) && stmpp_buffer->bytesused) {
    /* Returning data are always in plane 0 */
    stmpp_buffer->planes[0].bytesused = stmpp_buffer->bytesused;
    buffer->m.planes[0].bytesused = stmpp_buffer->bytesused;
  }

  buffer->timestamp.tv_sec = stmpp_buffer->timestamp / 1000000;
  buffer->timestamp.tv_usec = stmpp_buffer->timestamp % 1000000;

  buffer->flags = 0;
  if (stmpp_buffer_keyframe(stmpp_buffer))
    buffer->flags |= V4L2_BUF_FLAG_KEYFRAME;
  if (stmpp_buffer_error(stmpp_buffer)) buffer->flags |= V4L2_BUF_FLAG_ERROR;
  if (stmpp_buffer_queued(stmpp_buffer)) {
    buffer->flags |= V4L2_BUF_FLAG_QUEUED;
    if (stmpp_buffer_available(stmpp_buffer))
      buffer->flags |= V4L2_BUF_FLAG_DONE;
    else
      buffer->flags |= V4L2_BUF_FLAG_PREPARED;
  }

  buffer->field = V4L2_FIELD_NONE;
  memset(&buffer->timecode, 0, sizeof(buffer->timecode));
  buffer->sequence = 0;

  buffer->index = stmpp_buffer->index;

  LEAVE();
  return 0;
}

int stmpp_from_v4l2_buffer(struct stmpp_context *ctx,
                           struct v4l2_buffer *buffer,
                           struct stmpp_buffer *stmpp_buffer) {
  unsigned int i;

  ENTER();

  stmpp_buffer->length = buffer->length;

  ////copy data of each plane
  for (i = 0; i < buffer->length; i++) {
    stmpp_buffer->planes[i].length = buffer->m.planes[i].length;
    stmpp_buffer->planes[i].data_offset = buffer->m.planes[i].data_offset;
    stmpp_buffer->planes[i].bytesused = buffer->m.planes[i].bytesused;
    stmpp_buffer->planes[i].plane_size =
        stmpp_buffer->planes[i].bytesused - stmpp_buffer->planes[i].data_offset;

    stmpp_buffer->bytesused += stmpp_buffer->planes[i].plane_size;

    if (buffer->memory == V4L2_MEMORY_USERPTR)
      stmpp_buffer->planes[i].userptr = buffer->m.planes[i].m.userptr;
    else if (buffer->memory == V4L2_MEMORY_DMABUF)
      stmpp_buffer->planes[i].fd = buffer->m.planes[i].m.fd;
  }

  stmpp_buffer->timestamp =
      (uint64_t)buffer->timestamp.tv_sec * 1000000 + buffer->timestamp.tv_usec;

  LEAVE();
  return 0;
}
